from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aiohttp import ClientResponse, WSMessage

    from ._types import Json
    from .fortnite import AccountBoundMixin, BaseEntity, Recyclable, Upgradable


__all__ = (
    "EpicException",
    "HTTPException",
    "WSException",
    "WSConnectionError",
    "XMPPClosed",
    "FortniteException",
    "UnknownTemplateID",
    "BadItemAttributes",
    "ItemIsReadOnly",
    "ItemIsFavorited",
    "InvalidUpgrade",
)


class EpicException(Exception):
    pass


class HTTPException(EpicException):
    def __init__(self, response: ClientResponse, data: Json, /) -> None:
        self.response: ClientResponse = response
        self.data: Json = data.copy()

        _error_data = data if isinstance(data, dict) else {}
        self.server_code: str = _error_data.get(
            "errorCode", "unknown_error_code"
        )
        self.server_message: str = _error_data.get(
            "errorMessage", "An error occurred."
        )
        self.server_vars: list[str] = _error_data.get("messageVars", []).copy()

        self.originating_service: str | None = _error_data.get(
            "originatingService"
        )
        self.intent: str | None = _error_data.get("intent")

    def __str__(self) -> str:
        return f"{self.response.status} {self.response.reason} - {self.server_message}"


# TODO: Implement special methods on WS/XMPP Exception classes


class WSException(EpicException):

    def __init__(self, message: WSMessage, /) -> None:
        self.message: WSMessage = message


class WSConnectionError(WSException):
    pass


class XMPPClosed(WSException):
    pass


class FortniteException(EpicException):
    pass


class UnknownTemplateID(FortniteException):

    def __init__(self, item: BaseEntity, /) -> None:
        self.item: BaseEntity = item

    def __str__(self) -> str:
        return f"Unknown template ID: {self.item.template_id}"


class BadItemAttributes(FortniteException):

    def __init__(self, item: BaseEntity, /) -> None:
        self.item: BaseEntity = item

    def __str__(self) -> str:
        return f"Malformed/invalid item attributes: {self.item.raw_attributes}"


class ItemIsReadOnly(FortniteException):

    def __init__(self, item: AccountBoundMixin, /) -> None:
        self.item: AccountBoundMixin = item

    def __str__(self) -> str:
        return "Item can not be modified as it is not tied to a FullAccount"


class ItemIsFavorited(FortniteException):

    def __init__(self, item: Recyclable, /) -> None:
        self.item: Recyclable = item

    def __str__(self) -> str:
        return "Item is favorited so it can not be recycled"


class InvalidUpgrade(FortniteException):

    def __init__(self, item: Upgradable, /) -> None:
        self.item: Upgradable = item

    def __str__(self) -> str:
        return "An invalid target level/tier was specified"



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic

from pyepic._types import AccountT
from pyepic.errors import (
    BadItemAttributes,
    InvalidUpgrade,
    ItemIsFavorited,
    ItemIsReadOnly,
    UnknownTemplateID,
)
from pyepic.resources import lookup

from .base import AccountBoundMixin, BaseEntity

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable
    from typing import Any, ClassVar

    from pyepic._types import Attributes, Dict
    from pyepic.auth import AuthSession


__all__ = (
    "SaveTheWorldItem",
    "Recyclable",
    "Upgradable",
    "Schematic",
    "SchematicPerk",
    "SetBonusType",
    "SurvivorBase",
    "Survivor",
    "LeadSurvivor",
    "ActiveSetBonus",
    "FortStat",
    "SurvivorSquad",
)


class SaveTheWorldItem(BaseEntity):
    __slots__ = ("name", "type", "tier", "level", "rarity", "favorite")

    def __init__(
        self, template_id: str, raw_attributes: Attributes, /
    ) -> None:
        super().__init__(template_id, raw_attributes)

        items: dict[str, dict[str, str]] = lookup["Items"]

        for variation in (
            template_id,
            template_id[:-2] + "01",
            template_id.replace("Trap:tid", "Schematic:sid")[:-2] + "01",
            template_id.replace("Weapon:wid", "Schematic:sid")[:-2] + "01",
        ):
            if variation in items:
                lookup_id = variation
                break
        else:
            raise UnknownTemplateID(self)

        entry = items[lookup_id]

        self.name: str = entry["name"]
        self.type: str = lookup["ItemTypes"][entry["type"]]
        self.level: int = raw_attributes.get("level", 1)
        self.rarity: str = entry["rarity"].capitalize()
        self.favorite: bool = raw_attributes.get("favorite", False)

        tier = template_id[-1]
        try:
            self.tier = int(tier)
        except ValueError:
            self.tier = 1

    def __str__(self) -> str:
        return self.name


class Recyclable(
    Generic[AccountT], AccountBoundMixin[AccountT], SaveTheWorldItem
):
    __slots__ = ("account", "id")

    async def recycle(self, *, strict: bool = True) -> Dict:
        if self.favorite is True and strict is True:
            raise ItemIsFavorited(self)

        data = await self._auth_checker.mcp_operation(
            operation="RecycleItem",
            profile_id="campaign",
            json={"targetItemId": self.id},
        )

        self.account.uncache_stw_object(self)

        return data


class Upgradable(Generic[AccountT], Recyclable[AccountT]):
    __slots__ = ()

    __tier_mapping__: ClassVar[dict[int, str]] = {
        1: "i",
        2: "ii",
        3: "iii",
        4: "iv",
        5: "v",
    }

    async def upgrade(
        self, *, new_level: int, new_tier: int, conversion_index: int
    ) -> Dict:
        if new_tier not in range(self.tier, 6) or new_level not in range(
            self.level + 1, 61
        ):
            raise InvalidUpgrade(self)

        data = await self._auth_checker.mcp_operation(
            operation="UpgradeItemBulk",
            profile_id="campaign",
            json={
                "targetItemId": self.id,
                "desiredLevel": new_level,
                "desiredTier": self.__tier_mapping__[new_tier],
                "conversionRecipeIndexChoice": conversion_index,
            },
        )

        self.template_id = self.template_id.replace(
            f"_t0{self.tier}", f"_t0{new_tier}"
        )
        self.level, self.tier = new_level, new_tier
        if (
            isinstance(self, Schematic)
            and self.tier > 3
            and conversion_index == 1
        ):
            self.template_id = self.template_id.replace("_ore_", "_crystal_")

        return data


class Schematic(Generic[AccountT], Upgradable[AccountT]):
    __slots__ = ("perks",)

    def __init__(
        self,
        account: AccountT,
        item_id: str,
        template_id: str,
        raw_attributes: Attributes,
        /,
    ) -> None:
        try:
            super().__init__(account, item_id, template_id, raw_attributes)
        except UnknownTemplateID:
            super().__init__(
                account,
                item_id,
                template_id.replace("_crystal_", "_ore_"),
                raw_attributes,
            )
            self.template_id = template_id

        self.perks: tuple[SchematicPerk[AccountT], ...] = tuple(
            SchematicPerk(self, perk_id)
            for perk_id in raw_attributes.get("alterations", ())
        )

    @property
    def power_level(self) -> int:
        return lookup["ItemPowerLevels"]["Other"][self.rarity][str(self.tier)][
            str(self.level)
        ]


# TODO: find a way to implement perk description
# TODO: implement special methods
class SchematicPerk(Generic[AccountT]):
    __slots__ = ("schematic", "id", "rarity", "description")

    def __init__(
        self, schematic: Schematic[AccountT], perk_id: str, /
    ) -> None:
        self.schematic: Schematic[AccountT] = schematic
        self.id: str = perk_id

        try:
            self.rarity: str = (
                "Common",
                "Uncommon",
                "Rare",
                "Epic",
                "Legendary",
            )[int(perk_id[-1]) - 1]
        except (IndexError, ValueError):
            self.rarity: str = "Common"

        self.description: str = ...


@dataclass(kw_only=True, slots=True, frozen=True)
class SetBonusType:
    id: str
    name: str
    bonus: int
    requirement: int

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: SetBonusType, /) -> bool:
        return type(other) is type(self) and self.id == other.id


class SurvivorBase(Generic[AccountT], Upgradable[AccountT]):
    __slots__ = ("personality", "squad_id", "squad_index")

    def __init__(
        self,
        account: AccountT,
        item_id: str,
        template_id: str,
        raw_attributes: Attributes,
        /,
    ) -> None:
        super().__init__(account, item_id, template_id, raw_attributes)

        try:
            self.personality: str = raw_attributes["personality"].split(".")[
                -1
            ][2:]
            _index = raw_attributes["squad_slot_idx"]
        except KeyError:
            raise BadItemAttributes(self)

        self.squad_id: str | None = raw_attributes.get("squad_id") or None
        self.squad_index: int | None = _index if _index != -1 else None

    async def squad(
        self,
        auth_session: AuthSession | None = None,
        /,
        *,
        raise_read_only: bool = False,
        **kwargs: Any,
    ) -> SurvivorSquad[AccountT] | None:
        if self.squad_id is None:
            return

        try:
            return await self.account.survivor_squad_from_id(
                self.squad_id, auth_session or self._auth_checker, **kwargs
            )
        except ItemIsReadOnly as error:
            if raise_read_only is True:
                raise error
            raise ValueError("An authorization session needs to be passed")

    async def recycle(self, *, strict: bool = True) -> Dict:
        data = await super().recycle(strict=strict)

        squad = await self.squad(raise_read_only=True)
        squad.unslot_local(self)

        return data


class Survivor(Generic[AccountT], SurvivorBase[AccountT]):
    __slots__ = ("set_bonus_type",)

    def __init__(
        self,
        account: AccountT,
        item_id: str,
        template_id: str,
        raw_attributes: Attributes,
        /,
    ) -> None:
        super().__init__(account, item_id, template_id, raw_attributes)

        try:
            _set_bonus_type: str = (
                raw_attributes["set_bonus"]
                .split(".")[-1][2:]
                .replace("Low", "")
                .replace("High", "")
            )
            _set_bonus_data: dict[str, str | int] = lookup["SetBonuses"][
                _set_bonus_type
            ]
        except KeyError:
            raise BadItemAttributes(self)

        self.set_bonus_type: SetBonusType = SetBonusType(
            id=_set_bonus_type, **_set_bonus_data
        )

    @property
    def base_power_level(self) -> int:
        return lookup["ItemPowerLevels"]["Survivor"][self.rarity][
            str(self.tier)
        ][str(self.level)]


class LeadSurvivor(Generic[AccountT], SurvivorBase[AccountT]):
    __slots__ = ("preferred_squad_id",)

    def __init__(
        self,
        account: AccountT,
        item_id: str,
        template_id: str,
        raw_attributes: Attributes,
        /,
    ) -> None:
        super().__init__(account, item_id, template_id, raw_attributes)

        try:
            self.preferred_squad_id: str = lookup["LeadPreferredSquads"][
                raw_attributes["managerSynergy"]
            ]
        except KeyError:
            raise BadItemAttributes(self)

    @property
    def base_power_level(self) -> int:
        return lookup["ItemPowerLevels"]["LeadSurvivor"][self.rarity][
            str(self.tier)
        ][str(self.level)]


@dataclass(kw_only=True, slots=True, frozen=True)
class ActiveSetBonus(Generic[AccountT]):

    squad: SurvivorSquad[AccountT]
    set_bonus_type: SetBonusType

    def __str__(self) -> str:
        return str(self.set_bonus_type)

    def __eq__(self, other: ActiveSetBonus, /) -> bool:
        return (
            type(other) is type(self)
            and self.squad == other.squad
            and self.set_bonus_type == other.set_bonus_type
        )


@dataclass(kw_only=True, slots=True, frozen=True)
class FortStat:
    tech: int
    offense: int
    fortitude: int
    resistance: int

    def __add__(self, other: FortStat, /) -> FortStat:
        return FortStat(
            tech=self.tech + other.tech,
            offense=self.offense + other.offense,
            fortitude=self.fortitude + other.fortitude,
            resistance=self.resistance + other.resistance,
        )


class SurvivorSquad(Generic[AccountT], AccountBoundMixin[AccountT]):
    __slots__ = ("account", "id", "name", "lead_survivor", "survivors")

    def __init__(
        self,
        account: AccountT,
        item_id: str,
        /,
        *,
        lead_survivor: LeadSurvivor | None,
        survivors: Iterable[Survivor],
    ) -> None:
        super().__init__(account, item_id)

        self.name: str = lookup["SquadDetails"][self.id]["name"]
        self.lead_survivor: LeadSurvivor[AccountT] | None = lead_survivor

        slots = [None] * 7
        for survivor in survivors:
            slots[survivor.squad_index - 1] = survivor  # noqa

        self.survivors: tuple[Survivor[AccountT] | None, ...] = tuple(slots)

    def __str__(self) -> str:
        return self.name

    def __iter__(
        self,
    ) -> Generator[LeadSurvivor[AccountT] | Survivor[AccountT], None, None]:
        if self.lead_survivor is not None:
            yield self.lead_survivor
        yield from filter(
            lambda survivor: survivor is not None, self.survivors
        )

    def __contains__(self, survivor: SurvivorBase, /) -> bool:
        return survivor == self.lead_survivor or survivor in self.survivors

    @property
    def active_set_bonuses(self) -> list[ActiveSetBonus[AccountT]]:
        tally: dict[SetBonusType, int] = {}

        for survivor in self.survivors:
            if survivor is None:
                continue
            try:
                tally[survivor.set_bonus_type] += 1
            except KeyError:
                tally[survivor.set_bonus_type] = 1

        active_set_bonuses = []

        for set_bonus_type, count in tally.items():

            sets = count // set_bonus_type.requirement
            for _ in range(sets):
                active_set_bonus = ActiveSetBonus(
                    squad=self, set_bonus_type=set_bonus_type
                )
                active_set_bonuses.append(active_set_bonus)

        return active_set_bonuses

    @property
    def total_fort_stats(self) -> FortStat:
        fort_stats_data = dict(tech=0, offense=0, fortitude=0, resistance=0)

        count = 0

        lead = self.lead_survivor
        if lead is not None:
            pl = lead.base_power_level
            if lead.preferred_squad_id == self.id:
                count += pl * 2
            else:
                count += pl

        for survivor in self.survivors:
            if survivor is None:
                continue
            pl = survivor.base_power_level
            increments = lookup["LeadBonuses"]

            if lead is not None and lead.personality == survivor.personality:
                pl += increments[lead.rarity][0]
            elif lead is not None:
                pl += increments[lead.rarity][1]

            count += pl

        fort_type = lookup["SquadDetails"][self.id]["fort"]
        fort_stats_data[fort_type] += count

        return FortStat(**fort_stats_data)

    def slot_local(self, survivor: SurvivorBase, index: int, /) -> None:
        if index == 0:
            self.lead_survivor = survivor
        else:
            survivors = list(self.survivors)
            survivors[index - 1] = survivor
            self.survivors = tuple(survivors)

    def unslot_local(self, survivor: SurvivorBase, /) -> None:
        if survivor == self.lead_survivor:
            self.lead_survivor = None
        else:
            self.survivors = tuple(
                None if survivor == old_survivor else old_survivor
                for old_survivor in self.survivors
            )

    def unslot_all_local(self) -> None:
        self.lead_survivor = None
        self.survivors = None, None, None, None, None, None, None

    async def slot(self, survivor: SurvivorBase, index: int, /) -> Dict:
        previous_squad = await survivor.squad(raise_read_only=True)
        if previous_squad is not None:
            previous_squad.unslot_local(survivor)

        data = await self._auth_checker.mcp_operation(
            operation="AssignWorkerToSquad",
            profile_id="campaign",
            json={
                "characterId": survivor.id,
                "squadId": self.id,
                "slotIndex": index,
            },
        )

        self.slot_local(survivor, index)

        return data

    async def unslot(self, survivor: SurvivorBase, /) -> Dict:
        if survivor == self.lead_survivor:
            index = 0
        else:
            index = self.survivors.index(survivor) + 1

        data = await self._auth_checker.mcp_operation(
            operation="AssignWorkerToSquad",
            profile_id="campaign",
            json={"characterId": survivor.id, "squadId": "", "index": index},
        )

        self.unslot_local(survivor)

        return data

    async def unslot_all(self) -> Dict:
        data = await self._auth_checker.mcp_operation(
            operation="UnassignAllSquads",
            profile_id="campaign",
            json={"squadIds": [self.id]},
        )

        self.unslot_all_local()

        return data



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from logging import getLogger
from typing import TYPE_CHECKING, Generic

from pyepic.resources import lookup

from ._types import AuthT
from .errors import UnknownTemplateID
from .fortnite.base import AccountBoundMixin
from .fortnite.stw import LeadSurvivor, Schematic, Survivor, SurvivorSquad
from .route import FriendsService

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Coroutine
    from typing import Any, Self

    from ._types import Attributes, DCo, Dict, FriendType, List, STWItemT_co
    from .auth import AuthSession
    from .fortnite.stw import SaveTheWorldItem
    from .http import HTTPClient


__all__ = ("Friend", "PartialAccount", "FullAccount")


_logger = getLogger(__name__)


@dataclass(kw_only=True, slots=True, frozen=True)
class Friend(Generic[AuthT]):
    original: FullAccount[AuthT]
    account: PartialAccount
    type: FriendType

    created: datetime | None
    favorite: bool | None
    mutual: int | None
    alias: str | None
    note: str | None

    def __str__(self) -> str:
        return str(self.account)

    def __eq__(self, other: Friend, /) -> bool:
        return (
            isinstance(other, Friend)
            and self.original == other.original
            and self.account == other.account
            and self.type == other.type
        )


class PartialAccount:
    __slots__ = (
        "client",
        "data",
        "id",
        "display_name",
        "__raw_stw_data",
        "__stw_object_cache",
    )

    def __init__(self, client: HTTPClient, data: Dict, /) -> None:
        self.client: HTTPClient = client
        self.data: Dict = data
        self.id: str = data.get("id")
        self.display_name: str = data.get("displayName", "")

        self.__raw_stw_data: Dict | None = None
        self.__stw_object_cache: (
            dict[tuple[str, type[STWItemT_co]], list[STWItemT_co]] | None
        ) = None

    def __hash__(self) -> int:
        return hash(self.id)

    def __str__(self) -> str:
        return self.display_name

    def __eq__(self, other: PartialAccount, /) -> bool:
        return isinstance(other, PartialAccount) and self.id == other.id

    async def fetch_raw_stw_data(
        self, auth_session: AuthSession, /, *, use_cache: bool = True
    ) -> Dict:
        if self.__raw_stw_data is not None and use_cache is True:
            return self.__raw_stw_data

        data = await auth_session.mcp_operation(
            epic_id=self.id,
            path="public",
            operation="QueryPublicProfile",
            profile_id="campaign",
        )

        if self.client.cache_config.enable_mcp_caching is True:
            self.__raw_stw_data = data

        return data

    async def fetch_stw_objects(
        self,
        template_id_prefix: str,
        cls: type[STWItemT_co],
        auth_session: AuthSession,
        /,
        *,
        strict: bool = False,
        use_cache: bool = True,
    ) -> list[STWItemT_co]:
        key = template_id_prefix, cls
        cache = self.__stw_object_cache

        if cache is not None and use_cache is True and key in cache:
            return cache[key]

        data = await self.fetch_raw_stw_data(auth_session, use_cache=use_cache)
        items_data: Dict = data["profileChanges"][0]["profile"]["items"]

        items: list[STWItemT_co] = []

        item_data: Dict
        for item_id, item_data in items_data.items():
            template_id: str = item_data["templateId"]

            if not template_id.startswith(template_id_prefix):
                continue

            raw_attributes: Attributes = item_data["attributes"]

            if issubclass(cls, AccountBoundMixin):
                args = self, item_id, template_id, raw_attributes
            else:
                args = template_id, raw_attributes

            try:
                item = cls(*args)

            except UnknownTemplateID as error:
                if strict is True:
                    raise error

                _logger.error(error)
                continue

            items.append(item)

        if self.client.cache_config.enable_mcp_caching is True:
            if cache is None:
                self.__stw_object_cache = {key: items}
            else:
                cache[key] = items

        return items

    def uncache_stw_object(self, obj: SaveTheWorldItem, /) -> None:
        cls = type(obj)
        cache = self.__stw_object_cache

        if cache is None:
            return

        for key in cache:
            if key[1] is cls:
                try:
                    cache[key].remove(obj)
                except ValueError:
                    continue

    def schematics(
        self, auth_session: AuthSession, /, **kwargs: Any
    ) -> Coroutine[Any, Any, list[Schematic[Self]]]:
        return self.fetch_stw_objects(
            "Schematic:sid", Schematic, auth_session, **kwargs
        )

    def survivors(
        self, auth_session: AuthSession, /, **kwargs: Any
    ) -> Coroutine[Any, Any, list[Survivor[Self]]]:
        return self.fetch_stw_objects(
            "Worker:worker", Survivor, auth_session, **kwargs
        )

    def lead_survivors(
        self, auth_session: AuthSession, /, **kwargs: Any
    ) -> Coroutine[Any, Any, list[LeadSurvivor[Self]]]:
        return self.fetch_stw_objects(
            "Worker:manager", LeadSurvivor, auth_session, **kwargs
        )

    async def survivor_squads(
        self, auth_session: AuthSession, /, **kwargs: Any
    ) -> list[SurvivorSquad[Self]]:
        s1 = await self.survivors(auth_session, **kwargs)
        s2 = await self.lead_survivors(auth_session, **kwargs)
        survivors = s1 + s2

        mapping: dict = {
            squad_id: {"lead": None, "survivors": []}
            for squad_id in lookup["SquadDetails"]
        }

        for survivor in survivors:
            squad_id = survivor.squad_id
            if squad_id is None:
                continue

            elif isinstance(survivor, Survivor):
                mapping[squad_id]["survivors"].append(survivor)
            elif isinstance(survivor, LeadSurvivor):
                mapping[squad_id]["lead"] = survivor

        squads = []

        for squad_id, squad_composition in mapping.items():
            squad = SurvivorSquad(
                self,
                squad_id,
                lead_survivor=squad_composition["lead"],
                survivors=squad_composition["survivors"],
            )
            squads.append(squad)

        return squads

    async def survivor_squad_from_id(
        self, value: str, auth_session: AuthSession, /, **kwargs: Any
    ) -> SurvivorSquad[Self]:
        squads = await self.survivor_squads(auth_session, **kwargs)

        try:
            squad = next(
                squad
                for squad in squads
                if (squad.id == value or squad.name == value)
            )
        except StopIteration:
            raise ValueError("An invalid squad ID/name was passed.")

        return squad


class FullAccount(Generic[AuthT], PartialAccount):
    __slots__ = (
        "auth_session",
        "display_name_changes",
        "can_update_display_name",
        "first_name",
        "last_name",
        "country",
        "language",
        "email",
        "email_verified",
        "failed_login_attempts",
        "tfa_enabled",
        "last_login",
        "display_name_last_updated",
    )

    def __init__(self, auth_session: AuthT, data: Dict, /) -> None:
        super().__init__(auth_session.client, data)

        self.auth_session: AuthT = auth_session

        self.display_name_changes: int = data.get(
            "numberOfDisplayNameChanges", 0
        )
        self.can_update_display_name: bool | None = data.get(
            "canUpdateDisplayName"
        )

        self.first_name: str | None = data.get("name")
        self.last_name: str | None = data.get("lastName")
        self.country: str | None = data.get("country")
        self.language: str | None = data.get("preferredLanguage")

        if isinstance(self.language, str):
            self.language = self.language.capitalize()

        self.email: str | None = data.get("email")
        self.email_verified: bool | None = data.get("emailVerified")

        self.failed_login_attempts: int | None = data.get(
            "failedLoginAttempts"
        )
        self.tfa_enabled: bool | None = data.get("tfaEnabled")

        self.last_login: datetime = datetime.fromisoformat(
            data.get("lastLogin")
        )
        self.display_name_last_updated: datetime = datetime.fromisoformat(
            data.get("lastDisplayNameChange")
        )

    async def friends(
        self, *, friend_type: FriendType = "friends", use_cache: bool = True
    ) -> AsyncGenerator[Friend[AuthT], None]:
        route = FriendsService(
            "/friends/api/v1/{account_id}/summary", account_id=self.id
        )
        data: Dict = await self.auth_session.access_request("get", route)

        friend_type_data: List = data[friend_type]
        friend_type_data.sort(key=lambda entry: entry["accountId"])

        account_ids = tuple(entry["accountId"] for entry in friend_type_data)
        accounts = [
            account
            async for account in self.auth_session.fetch_accounts(
                *account_ids, use_cache=use_cache
            )
        ]
        # Fetching accounts doesn't necessarily preserve the order that IDs are passed in
        accounts.sort(key=lambda account: account.id)

        for i in range(len(friend_type_data)):
            entry = friend_type_data[i]

            try:
                created = datetime.fromisoformat(entry.get("created"))
            except (ValueError, TypeError):
                created = None

            yield Friend(
                original=self,
                account=accounts[i],
                type=friend_type,
                created=created,
                favorite=entry.get("favorite"),
                mutual=entry.get("mutual"),
                alias=entry.get("alias") or None,
                note=entry.get("note") or None,
            )

    def __friend(self, friend_id: str, op_type: str, op: str, /) -> DCo:
        route = FriendsService(
            "/friends/api/v1/{account_id}/{op_type}/{friend_id}",
            account_id=self.id,
            op_type=op_type,
            friend_id=friend_id,
        )
        return self.auth_session.access_request(op, route)

    def friend(self, account: PartialAccount, /) -> DCo:
        return self.__friend(account.id, "friends", "post")

    def unfriend(self, account: PartialAccount, /) -> DCo:
        return self.__friend(account.id, "friends", "delete")

    def block(self, account: PartialAccount, /) -> DCo:
        return self.__friend(account.id, "blocklist", "post")

    def unblock(self, account: PartialAccount, /) -> DCo:
        return self.__friend(account.id, "blocklist", "delete")



from __future__ import annotations

from abc import ABC
from typing import ClassVar
from urllib.parse import quote

__all__ = (
    "Route",
    "EpicGamesService",
    "AccountService",
    "FriendsService",
    "MCPService",
)


class Route(ABC):
    BASE: ClassVar[str] = ""

    __slots__ = ("__path", "__kwargs")

    def __init__(self, path: str, /, **kwargs: str) -> None:
        if self.BASE == "":
            raise ValueError("Route must have a base.")

        self.__path: str = path
        self.__kwargs: dict[str, str] = {
            k: self.__quote__(v) for k, v in kwargs.items()
        }

    def __hash__(self) -> int:
        return hash(self.url)

    def __str__(self) -> str:
        return self.url

    def __eq__(self, other: Route, /) -> bool:
        return isinstance(other, Route) and self.url == other.url

    @staticmethod
    def __quote__(string: str, /) -> str:
        string = quote(string)
        string = string.replace("/", "%2F")
        return string

    @property
    def url(self) -> str:
        return self.BASE + self.__path.format(**self.__kwargs)


class EpicGamesService(Route):
    __slots__ = ()

    BASE = "https://www.epicgames.com"


class AccountService(Route):
    __slots__ = ()

    BASE = "https://account-public-service-prod.ol.epicgames.com"


class FriendsService(Route):
    __slots__ = ()

    BASE = "https://friends-public-service-prod.ol.epicgames.com"


class MCPService(Route):
    __slots__ = ()

    BASE = "https://fngw-mcp-gc-livefn.ol.epicgames.com"



from __future__ import annotations

from collections.abc import Coroutine
from typing import TYPE_CHECKING, Any, Literal, TypedDict, TypeVar

from .route import Route

if TYPE_CHECKING:
    from .account import PartialAccount
    from .auth import AuthSession
    from .fortnite import SaveTheWorldItem

    STWItemT_co = TypeVar(
        "STWItemT_co", covariant=True, bound=SaveTheWorldItem
    )

    AuthT = TypeVar("AuthT", bound=AuthSession)
    AccountT = TypeVar("AccountT", bound=PartialAccount)
else:
    AuthT = TypeVar("AuthT", bound="AuthSession")
    AccountT = TypeVar("AccountT", bound="PartialAccount")


URL = Route | str

Dict = dict[str, Any]
List = list[Dict]
Json = Dict | List

DCo = Coroutine[Any, Any, Dict]
JCo = Coroutine[Any, Any, Json]

Attributes = dict[str, Any]

FriendType = Literal[
    "friends", "incoming", "outgoing", "suggested", "blocklist"
]


class PartialCacheEntry(TypedDict):
    account: PartialAccount
    expires: float



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Generic

from pyepic._types import AccountT
from pyepic.errors import ItemIsReadOnly

if TYPE_CHECKING:
    from typing import Any

    from pyepic._types import Attributes
    from pyepic.auth import AuthSession


__all__ = ("AccountBoundMixin", "BaseEntity")


class AccountBoundMixin(ABC, Generic[AccountT]):
    # Defer to subclasses
    __slots__ = ()

    def __init__(
        self, account: AccountT, item_id: str, *args: Any, **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)

        self.account: AccountT = account  # noqa
        self.id: str = item_id  # noqa

    def __eq__(self, other: AccountBoundMixin, /) -> bool:
        return (
            type(other) is type(self)
            and self.account == other.account
            and self.id == other.id
        )

    @property
    def _auth_checker(self) -> AuthSession:
        try:
            return self.account.auth_session
        except AttributeError:
            raise ItemIsReadOnly(self)


class BaseEntity(ABC):
    __slots__ = ("template_id", "raw_attributes")

    def __init__(
        self, template_id: str, raw_attributes: Attributes, /
    ) -> None:
        self.template_id: str = template_id
        self.raw_attributes: Attributes = raw_attributes



from __future__ import annotations

from asyncio import create_task, get_running_loop, sleep
from base64 import b64encode
from collections import OrderedDict
from dataclasses import dataclass
from logging import getLogger
from time import time
from typing import TYPE_CHECKING

from aiohttp import ClientResponseError, ClientSession
from aiohttp.helpers import sentinel

from .auth import AuthManager, AuthSession
from .errors import HTTPException
from .route import AccountService, EpicGamesService

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop, Task
    from types import TracebackType
    from typing import Any, Self

    from aiohttp import BaseConnector, ClientResponse, ClientTimeout

    from ._types import URL, AuthT, DCo, JCo, Json, PartialCacheEntry
    from .account import PartialAccount


__all__ = ("HTTPRetryConfig", "CacheConfig", "XMPPConfig", "HTTPClient")


_logger = getLogger(__name__)


@dataclass(kw_only=True, slots=True, frozen=True)
class HTTPRetryConfig:
    max_retries: int = 5
    max_wait_time: float = 65.0

    handle_ratelimits: bool = True
    max_retry_after: float = 60.0

    handle_backoffs: bool = True
    backoff_factor: float = 1.5
    backoff_start: float = 1.0
    backoff_cap: float = 20


@dataclass(kw_only=True, slots=True, frozen=True)
class CacheConfig:
    enable_partial_caching: bool = True
    partial_cache_max_size: int = 1000
    partial_cache_max_age: float = 900.0
    partial_cache_prune_interval: float = 60.0

    enable_full_caching: bool = True
    full_cache_max_age: float = 900.0

    enable_mcp_caching: bool = True


@dataclass(kw_only=True, slots=True, frozen=True)
class XMPPConfig:

    xml_version: str = "1.0"
    xmpp_version: str = "1.0"

    domain: str = "xmpp-service-prod.ol.epicgames.com"
    host: str = "prod.ol.epicgames.com"
    port: int = 443

    platform: str = "WIN"

    connect_timeout: float = 10.0
    ping_interval: float = 60.0
    stop_timeout: float = 1.0


class HTTPClient:
    __slots__ = (
        "__loop",
        "retry_config",
        "__cache_config",
        "__connector",
        "__timeout",
        "__session",
        "client_id",
        "client_secret",
        "__partial_cache",
        "__prune_task",
        "__xmpp_config",
    )

    def __init__(
        self,
        *,
        loop: AbstractEventLoop | None = None,
        retry_config: HTTPRetryConfig | None = None,
        cache_config: CacheConfig | None = None,
        xmpp_config: XMPPConfig | None = None,
        connector: BaseConnector | None = None,
        timeout: ClientTimeout | None = None,
    ) -> None:
        self.__loop: AbstractEventLoop = loop or get_running_loop()
        self.retry_config: HTTPRetryConfig = retry_config or HTTPRetryConfig()
        self.__cache_config: CacheConfig = cache_config or CacheConfig()
        self.__xmpp_config: XMPPConfig = xmpp_config or XMPPConfig()

        self.__connector: BaseConnector | None = connector
        self.__timeout: ClientTimeout | None = timeout

        self.__session: ClientSession | None = None

        self.client_id: str = "ec684b8c687f479fadea3cb2ad83f5c6"
        self.client_secret: str = "e1f31c211f28413186262d37a13fc84d"

        self.__partial_cache: OrderedDict[str, PartialCacheEntry] | None = None
        self.__prune_task: Task | None = None

    async def __aenter__(self) -> Self:
        await self.create_connection()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.close_connection()

    @property
    def loop(self) -> AbstractEventLoop:
        return self.__loop

    @property
    def cache_config(self) -> CacheConfig:
        return self.__cache_config

    @property
    def xmpp_config(self) -> XMPPConfig:
        return self.__xmpp_config

    @property
    def connector(self) -> BaseConnector:
        return self.__connector

    @property
    def is_open(self) -> bool:
        return self.__session is not None and not self.__session.closed

    async def create_connection(self) -> None:
        self.__session = ClientSession(
            connector=self.__connector,
            connector_owner=self.__connector is None,
            timeout=self.__timeout or sentinel,
        )
        if self.__cache_config.enable_partial_caching is True:
            self.__setup_cache()

    async def close_connection(self) -> None:
        if self.is_open is True:
            await self.__session.close()
        if self.__cache_config.enable_partial_caching is True:
            self.__destroy_cache()

    @staticmethod
    async def response_to_json(response: ClientResponse, /) -> Json:
        try:
            return await response.json()
        except ClientResponseError:
            # This should only happen if we receive an empty response from Epic Games.
            _logger.debug(
                "Failed to decode response payload from %s %s",
                response.method.upper(),
                response.url,
            )
            return {}

    @staticmethod
    def get_retry_after(error: HTTPException, /) -> float | None:
        retry_after = error.response.headers.get("Retry-After")
        if retry_after is not None:
            return float(retry_after)

        try:
            return float(error.server_vars[0])
        except (IndexError, TypeError, ValueError):
            return

    async def make_request(
        self, method: str, raw_url: str, /, **kwargs: Any
    ) -> Json:
        if self.is_open is False:
            raise RuntimeError("HTTP session is closed.")

        pre_time = time()
        async with self.__session.request(
            method, raw_url, **kwargs
        ) as response:
            _logger.debug(
                "%s %s returned %s %s in %.3fs",
                method.upper(),
                raw_url,
                response.status,
                response.reason,
                time() - pre_time,
            )

            data = await self.response_to_json(response)

            if 200 <= response.status < 400:
                return data

            raise HTTPException(response, data)

    async def request(self, method: str, url: URL, /, **kwargs: Any) -> Json:
        url = str(url)
        config = self.retry_config

        tries = 0
        total_slept = 0
        backoff = config.backoff_start

        while True:
            tries += 1
            sleep_time = 0

            try:
                return await self.make_request(method, url, **kwargs)

            except HTTPException as error:
                if tries >= config.max_retries:
                    raise error

                if (
                    error.server_code
                    == "errors.com.epicgames.common.throttled"
                    or error.response.status == 429
                ):
                    retry_after = self.get_retry_after(error)

                    if retry_after is not None:
                        if (
                            config.handle_ratelimits is True
                            and retry_after <= config.max_retry_after
                        ):
                            sleep_time = retry_after

                    else:
                        backoff *= config.backoff_factor
                        if (
                            config.handle_backoffs is True
                            and backoff <= config.backoff_cap
                        ):
                            sleep_time = backoff

                elif (
                    error.server_code
                    == "errors.com.epicgames.common.server_error"
                    or error.server_code
                    == "errors.com.epicgames.common.concurrent_modification_error"
                    or error.response.status >= 500
                ):
                    sleep_time = 2 * (tries - 1) + 0.5

                if sleep_time > 0:
                    total_slept += sleep_time
                    if total_slept > config.max_wait_time:
                        raise error

                    _logger.debug(
                        "Retrying %s %s in %.3fs...",
                        method.upper(),
                        url,
                        sleep_time,
                    )

                    await sleep(sleep_time)
                    continue

                raise error

    def get(self, url: URL, /, **kwargs: Any) -> JCo:
        return self.request("get", url, **kwargs)

    def put(self, url: URL, /, **kwargs: Any) -> JCo:
        return self.request("put", url, **kwargs)

    def post(self, url: URL, /, **kwargs: Any) -> JCo:
        return self.request("post", url, **kwargs)

    def patch(self, url: URL, /, **kwargs: Any) -> JCo:
        return self.request("patch", url, **kwargs)

    def delete(self, url: URL, /, **kwargs: Any) -> JCo:
        return self.request("delete", url, **kwargs)

    @property
    def user_auth_path(self) -> EpicGamesService:
        return EpicGamesService(
            "/id/api/redirect?clientId={client_id}&responseType=code",
            client_id=self.client_id,
        )

    @property
    def auth_exchange_path(self) -> AccountService:
        return AccountService("/account/api/oauth/token")

    @property
    def auth_exchange_secret(self) -> str:
        return b64encode(
            f"{self.client_id}:{self.client_secret}".encode()
        ).decode()

    def create_auth_session(
        self,
        auth_code: str,
        /,
        *,
        cls: type[AuthT] = AuthSession,
        start_xmpp: bool = True,
    ) -> AuthManager[AuthT]:
        if not issubclass(cls, AuthSession):
            raise TypeError("Class should be a subclass of AuthSession")

        request_coro: DCo = self.post(
            self.auth_exchange_path,
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "Authorization": f"basic {self.auth_exchange_secret}",
            },
            data={"grant_type": "authorization_code", "code": auth_code},
        )

        return AuthManager(self, request_coro, cls=cls, start_xmpp=start_xmpp)

    def renew_auth_session(self, refresh_token: str, /) -> DCo:
        return self.post(
            self.auth_exchange_path,
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "Authorization": f"basic {self.auth_exchange_secret}",
            },
            data={
                "grant_type": "refresh_token",
                "refresh_token": refresh_token,
            },
        )

    def __setup_cache(self) -> None:
        self.__partial_cache = OrderedDict()
        self.__prune_task = create_task(self.__prune_cache())

    def __destroy_cache(self) -> None:
        task = self.__prune_task
        if task is not None and task.done() is False:
            task.cancel()
        self.__partial_cache = None

    async def __prune_cache(self) -> None:
        while True:
            await sleep(self.__cache_config.partial_cache_prune_interval)

            expired_accounts = []

            for entry in self.__partial_cache.values():
                account = entry["account"]
                if (
                    entry["expires"] < time()
                    and account not in expired_accounts
                ):
                    expired_accounts.append(account)

            for account in expired_accounts:
                self.uncache_account(account)

    def get_account(self, lookup: str, /) -> PartialAccount | None:
        if self.__cache_config.enable_partial_caching is False:
            return

        cache = self.__partial_cache
        entry = cache.get(lookup)

        if entry is not None:
            account = entry.get("account")
            cache.move_to_end(account.id)
            cache.move_to_end(account.display_name)
            return account

    def cache_account(self, account: PartialAccount, /) -> None:
        if self.__cache_config.enable_partial_caching is False:
            return
        # Reject accounts with incomplete lookup data
        # Avoids accidentally overwriting cached accounts
        elif account.display_name == "" or account.id is None:
            return

        config = self.__cache_config
        cache = self.__partial_cache

        new_entry = {
            "account": account,
            "expires": time() + config.partial_cache_max_age,
        }
        for key in account.id, account.display_name:
            cache[key] = new_entry
            cache.move_to_end(key)

        while len(cache) * 2 > config.partial_cache_max_size:
            for _ in range(2):
                cache.popitem(False)

    def uncache_account(self, account: PartialAccount, /) -> None:
        if self.__cache_config.enable_partial_caching is False:
            return

        cache = self.__partial_cache
        cache.pop(account.id, None)
        cache.pop(account.display_name, None)



from __future__ import annotations

from asyncio import Event, create_task, sleep, wait_for
from base64 import b64encode, urlsafe_b64encode
from logging import getLogger
from random import getrandbits
from traceback import print_exception
from typing import TYPE_CHECKING
from uuid import uuid4
from xml.etree.ElementTree import XMLPullParser

from aiohttp import ClientSession, WSMsgType

from .errors import WSConnectionError, XMPPClosed

if TYPE_CHECKING:
    from asyncio import Task
    from collections.abc import Coroutine, Iterable
    from typing import Any
    from xml.etree.ElementTree import Element

    from aiohttp import ClientWebSocketResponse, WSMessage

    from .auth import AuthSession
    from .http import XMPPConfig

    SendCoro = Coroutine[Any, Any, None]


if __import__("sys").version_info <= (3, 11):
    from asyncio import TimeoutError


__all__ = (
    "XMLNamespaces",
    "Stanza",
    "XMLGenerator",
    "XMLProcessor",
    "XMPPWebsocketClient",
)


_logger = getLogger(__name__)


class XMLNamespaces:

    CTX = "jabber:client"
    STREAM = "http://etherx.jabber.org/streams"
    SASL = "urn:ietf:params:xml:ns:xmpp-sasl"
    BIND = "urn:ietf:params:xml:ns:xmpp-bind"
    PING = "urn:xmpp:ping"


class Stanza:
    __slots__ = ("_tag", "_text", "_children", "_attributes", "_id")

    def __init__(
        self,
        *,
        tag: str,
        text: str = "",
        children: Iterable[Stanza] = (),
        **attributes: str,
    ) -> None:
        if text and children:
            raise ValueError("Invalid combination of Stanza arguments passed")
        elif attributes.get("id"):
            _logger.warning(
                "Stanza.__init__ received an 'ID' keyword argument"
            )

        self._tag = tag
        self._text = text
        self._children = tuple(children)
        self._attributes = attributes
        self._id = self._attributes["id"] = self.new_id()

    def __str__(self) -> str:
        attrs_str = ""
        for key, value in self._attributes.items():
            attrs_str += f" {key}='{value}'"
        if self._text:
            return f"<{self._tag}{attrs_str}>{self._text}</{self._tag}>"
        elif self._children:
            return f"<{self._tag}{attrs_str}>{''.join(str(child) for child in self._children)}</{self._tag}>"
        else:
            return f"<{self._tag}{attrs_str}/>"

    def __eq__(self, other: Stanza | str, /) -> bool:
        return str(self) == str(other)

    @property
    def id(self) -> str:
        return self._id

    @staticmethod
    def new_id():
        # Credit: aioxmpp
        _id = getrandbits(120)
        _id = _id.to_bytes((_id.bit_length() + 7) // 8, "little")
        _id = urlsafe_b64encode(_id).rstrip(b"=").decode("ascii")
        return ":" + _id


class XMLGenerator:
    __slots__ = ("xmpp",)

    def __init__(self, xmpp: XMPPWebsocketClient, /) -> None:
        self.xmpp: XMPPWebsocketClient = xmpp

    @property
    def uuid(self):
        return uuid4().hex.upper()

    @property
    def xml_prolog(self) -> str:
        return f"<?xml version='{self.xmpp.config.xml_version}'?>"

    @property
    def open(self) -> str:
        return (
            f"<stream:stream xmlns='{XMLNamespaces.CTX}' "
            f"xmlns:stream='{XMLNamespaces.STREAM}' "
            f"to='{self.xmpp.config.host}' "
            f"version='{self.xmpp.config.xmpp_version}'>"
        )

    @property
    def quit(self) -> str:
        return "</stream:stream>"

    @property
    def b64_plain(self) -> str:
        acc_id = self.xmpp.auth_session.account_id
        acc_tk = self.xmpp.auth_session.access_token
        return b64encode(f"\x00{acc_id}\x00{acc_tk}".encode()).decode()

    def auth(self, mechanism: str, /) -> Stanza:
        if mechanism == "PLAIN":
            auth = self.b64_plain
        else:
            # Expected authorization mechanism is PLAIN
            # But implement other mechanisms here if needed
            raise NotImplementedError
        return Stanza(
            tag="auth",
            text=auth,
            xmlns=XMLNamespaces.SASL,
            mechanism=mechanism,
        )

    @staticmethod
    def ping() -> Stanza:
        child = Stanza(tag="ping", xmlns=XMLNamespaces.PING)
        return Stanza(tag="iq", type="get", children=(child,))


class XMLProcessor:
    __slots__ = (
        "xmpp",
        "generator",
        "parser",
        "root",
        "outbound_ids",
        "xml_depth",
    )

    def __init__(self, xmpp: XMPPWebsocketClient, /) -> None:
        self.xmpp: XMPPWebsocketClient = xmpp
        self.generator: XMLGenerator = XMLGenerator(self.xmpp)

        self.parser: XMLPullParser | None = None
        self.root: Element | None = None
        self.outbound_ids: list[str] = []
        self.xml_depth: int = 0

    def setup(self) -> None:
        self.parser = XMLPullParser(("start", "end"))

    def teardown(self) -> None:
        self.parser = None
        self.outbound_ids = []
        self.xml_depth = 0
        self.root = None

    def process(self, message: WSMessage, /) -> None:
        if self.parser is None:
            raise RuntimeError("XML parser doesn't exist")

        self.parser.feed(message.data)

        # Inspiration: slixmpp
        event: str
        xml: Element
        for event, xml in self.parser.read_events():

            if event == "start":
                if self.xml_depth == 0:
                    self.root = xml
                self.xml_depth += 1

            elif event == "end":
                self.xml_depth -= 1

                if self.xml_depth == 0:
                    raise XMPPClosed(message)

                elif self.xml_depth == 1:
                    self.root.clear()


class XMPPWebsocketClient:
    __slots__ = (
        "auth_session",
        "config",
        "session",
        "ws",
        "processor",
        "recv_task",
        "ping_task",
        "cleanup_event",
        "exceptions",
    )

    def __init__(self, auth_session: AuthSession, /) -> None:
        self.auth_session: AuthSession = auth_session
        self.config: XMPPConfig = auth_session.client.xmpp_config

        self.session: ClientSession | None = None
        self.ws: ClientWebSocketResponse | None = None

        self.processor: XMLProcessor = XMLProcessor(self)

        self.recv_task: Task | None = None
        self.ping_task: Task | None = None
        self.cleanup_event: Event | None = None

        self.exceptions: list[Exception] = []

    @property
    def running(self) -> bool:
        return self.ws is not None and not self.ws.closed

    @property
    def most_recent_exception(self) -> Exception | None:
        try:
            return self.exceptions[-1]
        except IndexError:
            return None

    async def send(
        self, source: Stanza | str, /, *, with_xml_prolog: bool = False
    ) -> None:
        if isinstance(source, Stanza):
            self.processor.outbound_ids.append(source.id)
            source = str(source)
        if with_xml_prolog is True:
            source = self.processor.generator.xml_prolog + source

        await self.ws.send_str(source)
        self.auth_session.action_logger(f"SENT: {source}")

    async def ping_loop(self) -> None:
        while True:
            await sleep(self.config.ping_interval)
            await self.send(self.processor.generator.ping())

    async def recv_loop(self) -> None:
        self.auth_session.action_logger("Websocket receiver running")

        try:
            while True:
                message = await self.ws.receive()

                if message.type == WSMsgType.TEXT:
                    self.auth_session.action_logger(f"RECV: {message.data}")
                    self.processor.process(message)

                else:
                    raise WSConnectionError(message)

        except Exception as exception:
            if isinstance(exception, XMPPClosed):
                txt = "Websocket received closing message"
                level = _logger.debug
                print_exc = False
            else:
                txt = "Websocket encountered a fatal error"
                level = _logger.error
                print_exc = True

            self.auth_session.action_logger(txt, level=level)
            self.exceptions.append(exception)

            if print_exc is True:
                print_exception(exception)

            create_task(self.cleanup())  # noqa

        finally:
            self.auth_session.action_logger("Websocket receiver stopped")

    async def start(self) -> None:
        if self.running is True:
            return

        http = self.auth_session.client
        xmpp = self.config

        self.session = ClientSession(
            connector=http.connector, connector_owner=http.connector is None
        )
        self.ws = await self.session.ws_connect(
            f"wss://{xmpp.domain}:{xmpp.port}",
            timeout=xmpp.connect_timeout,
            protocols=("xmpp",),
        )
        self.processor.setup()

        self.recv_task = create_task(self.recv_loop())
        self.ping_task = create_task(self.ping_loop())
        self.cleanup_event = Event()

        self.auth_session.action_logger("XMPP started")

        # Let one iteration of the event loop pass
        # Before sending our opening message
        # So the receiver can initialise first
        await sleep(0)
        await self.send(self.processor.generator.open, with_xml_prolog=True)

    async def stop(self) -> None:
        if self.running is False:
            return

        await self.send(self.processor.generator.quit)

        try:
            await wait_for(self.wait_for_cleanup(), self.config.stop_timeout)
        except TimeoutError:
            await self.cleanup()

    async def wait_for_cleanup(self) -> None:
        if self.cleanup_event is None:
            return
        await self.cleanup_event.wait()

    async def cleanup(self) -> None:
        self.recv_task.cancel()
        self.ping_task.cancel()
        self.cleanup_event.set()

        await self.ws.close()
        await self.session.close()

        self.session = None
        self.ws = None
        self.processor.teardown()

        self.recv_task = None
        self.ping_task = None
        self.cleanup_event = None

        self.auth_session.action_logger("XMPP stopped")



from __future__ import annotations

from datetime import datetime
from logging import getLogger
from time import time
from typing import TYPE_CHECKING, Generic

from ._types import AuthT
from .account import FullAccount, PartialAccount
from .errors import HTTPException
from .route import AccountService, MCPService
from .utils import utc_now
from .xmpp import XMPPWebsocketClient

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable, Generator
    from types import TracebackType
    from typing import Any, Self

    from ._types import DCo, Dict, JCo, Json, List
    from .http import HTTPClient
    from .route import Route


__all__ = ("AuthManager", "AuthSession")


_logger = getLogger(__name__)


class AuthManager(Generic[AuthT]):
    __slots__ = (
        "__client",
        "__request_coro",
        "__cls",
        "__start_xmpp",
        "__auth_session",
    )

    def __init__(
        self,
        client: HTTPClient,
        request_coro: DCo,
        /,
        *,
        cls: type[AuthT],
        start_xmpp: bool,
    ) -> None:
        self.__client: HTTPClient = client
        self.__request_coro: DCo = request_coro
        self.__cls: type[AuthT] = cls
        self.__start_xmpp: bool = start_xmpp
        self.__auth_session: AuthT | None = None

    def __await__(self) -> Generator[Any, None, AuthT]:
        return self.__construct__().__await__()

    async def __aenter__(self) -> AuthT:
        return await self.__construct__()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.__auth_session.kill(stop_xmpp=self.__start_xmpp)

    async def __construct__(self) -> AuthT:
        data = await self.__request_coro
        self.__auth_session: AuthT = self.__cls(self.__client, data)
        if self.__start_xmpp is True:
            await self.__auth_session.xmpp.start()
        return self.__auth_session


class AuthSession:
    __slots__ = (
        "data",
        "client",
        "account_id",
        "access_token",
        "refresh_token",
        "access_expires",
        "refresh_expires",
        "_killed",
        "__cached_account",
        "__cached_account_expires",
        "xmpp",
    )

    def __init__(self, client: HTTPClient, data: Dict, /) -> None:
        self.client: HTTPClient = client

        self._renew_data(data)
        self.action_logger("initialised")

        self._killed = False

        self.__cached_account: FullAccount[Self] | None = None
        self.__cached_account_expires: float | None = None

        self.xmpp: XMPPWebsocketClient = XMPPWebsocketClient(self)

    def _renew_data(self, data: Dict, /) -> None:
        self.data: Dict = data
        self.account_id: str = data.get("account_id")
        self.access_token: str = data.get("access_token")
        self.refresh_token: str = data.get("refresh_token")
        self.access_expires: datetime = datetime.fromisoformat(
            data.get("expires_at")
        )
        self.refresh_expires: datetime = datetime.fromisoformat(
            data.get("refresh_expires_at")
        )

    def action_logger(
        self, action: str, /, *, level: Callable[..., None] = _logger.debug
    ) -> None:
        level(
            "Auth session %s %s. (Account ID: %s)",
            self.access_token,
            action,
            self.account_id,
        )

    @property
    def active(self) -> bool:
        return not self._killed and self.access_expires > utc_now()

    @property
    def expired(self) -> bool:
        return self._killed or self.refresh_expires < utc_now()

    @property
    def auth_headers(self) -> dict[str, str]:
        return {"Authorization": f"bearer {self.access_token}"}

    async def renew(self, *, force: bool = False) -> Dict | None:
        if self.active is True and force is False:
            return

        data = await self.client.renew_auth_session(self.refresh_token)

        self._renew_data(data)
        self.action_logger("renewed")

        return data

    async def access_request(
        self, method: str, route: Route, /, **kwargs: Any
    ) -> Json:
        headers = kwargs.pop("headers", None) or self.auth_headers

        try:
            return await self.client.request(
                method, route, headers=headers, **kwargs
            )
        except HTTPException as error:
            if error.response.status != 401 or self.expired is True:
                raise error

            await self.renew()
            return await self.client.request(
                method, route, headers=headers, **kwargs
            )

    async def kill(
        self, *, stop_xmpp: bool = True, force: bool = False
    ) -> None:
        if stop_xmpp is True:
            await self.xmpp.stop()

        if self.expired is True and force is False:
            return

        route = AccountService(
            "/account/api/oauth/sessions/kill/{access_token}",
            access_token=self.access_token,
        )
        try:
            await self.access_request("delete", route)
        except HTTPException:
            pass

        self._killed = True
        self.action_logger("killed")

    async def fetch_account(
        self,
        *,
        account_id: str | None = None,
        display_name: str | None = None,
        use_cache: bool = True,
    ) -> PartialAccount:
        lookup = account_id or display_name

        if lookup is None:
            raise ValueError("An account ID or display name is required")

        elif use_cache is True:
            account = self.client.get_account(lookup)
            if account is not None:
                return account

        if account_id is not None:
            route = AccountService(
                "/account/api/public/account/{account_id}",
                account_id=account_id,
            )
        else:
            route = AccountService(
                "/account/api/public/account/displayName/{display}",
                display=display_name,
            )

        data: Dict = await self.access_request("get", route)

        account = PartialAccount(self.client, data)
        self.client.cache_account(account)

        return account

    async def fetch_accounts(
        self, *account_ids: str, use_cache: bool = True
    ) -> AsyncGenerator[PartialAccount, None]:
        account_ids = tuple(set(account_ids))

        if use_cache is True:
            needs_fetching = []

            for account_id in account_ids:
                account = self.client.get_account(account_id)

                if account is not None:
                    yield account
                else:
                    needs_fetching.append(account_id)

        else:
            needs_fetching = account_ids

        if needs_fetching:
            chunks = [
                needs_fetching[i : i + 100]
                for i in range(0, len(needs_fetching), 100)
            ]
            route = AccountService("/account/api/public/account")

            for chunk in chunks:
                data: List = await self.access_request(
                    "get",
                    route,
                    params=[("accountId", account_id) for account_id in chunk],
                )

                for entry in data:
                    account = PartialAccount(self.client, entry)
                    self.client.cache_account(account)

                    yield account

    def __cache_self(self, account: FullAccount, /) -> None:
        self.__cached_account = account
        self.__cached_account_expires = (
            time() + self.client.cache_config.full_cache_max_age
        )

    async def __fetch_self(self) -> FullAccount[Self]:
        route = AccountService(
            "/account/api/public/account/{account_id}",
            account_id=self.account_id,
        )
        data: Dict = await self.access_request("get", route)
        return FullAccount(self, data)

    async def account(self, *, use_cache: bool = True) -> FullAccount[Self]:
        if (
            use_cache is True
            and self.__cached_account is not None
            and self.__cached_account_expires > time()
        ):
            return self.__cached_account

        account = await self.__fetch_self()
        if self.client.cache_config.enable_full_caching is True:
            self.__cache_self(account)
        return account

    def mcp_operation(
        self,
        *,
        method: str = "post",
        epic_id: str | None = None,
        path: str = "client",
        operation: str = "QueryProfile",
        profile_id: str = "athena",
        json: Dict | None = None,
    ) -> JCo:
        epic_id = epic_id or self.account_id
        json = json or {}

        route = MCPService(
            "/fortnite/api/game/v2/profile/{account_id}/{path}/{operation}?profileId={profile_id}",
            account_id=epic_id,
            path=path,
            operation=operation,
            profile_id=profile_id,
        )

        return self.access_request(method, route, json=json)
